      subroutine cjsolve(kkase,xmass_up,temp_up,den_up,mach, &
                         qx,xmass_det,ener_up,pres_up,cs_up, &
                         vel_det,vel_x,temp_x,den_x,ener_x,pres_x,cs_x)

      include 'implno.dek'
      include 'const.dek'
      include 'vector_eos.dek'
      include 'network.dek'

! solves the hugoniot and rayleigh relations for a detonation or a shock.
! an nse distribution is assumed for the detonated material.

! input:
! kkase     = 1 for a chapman-jouget detonation
!           = 2 for a strong point driven detonation
!           = 3 for a weak point driven detonation
!           = 4 for a shock wave
! xmass_up  = upstream composition vector
! temp_up   = temperature of upstream material
! den_up    = density of upstream material
! mach      = mach number of shock or detonation


! output:
! qx        = energy/gram from burning
! xmass_det = burned composition
! ener_up   = energy of upstream material
! pres_up   = pressure of upstream material
! cs_up     = sound speed of upstream material
! vel_det   = speed of detonation or shock front = upstream speed
! vel_x     = speed behind detonation or shock = downstream speed
! temp_x    = temperature of downstream material
! den_x     = density of downstream material
! ener_x    = internal energy of downstream material
! pres_x    = pressure of downstream material
! cs_x      = sound speed of downstream material


! declare the pass
      integer  :: kkase
      real*8   :: xmass_up(1),temp_up,den_up,mach,qx,xmass_det(1), &
                  ener_up,pres_up,cs_up,vel_det,vel_x,temp_x,den_x, &
                  ener_x,pres_x,cs_x


! common block communication with the routine cjfunc
      integer    ::    kase
      real*8     ::    xmup(abignet),den1,temp1,p1,u1,v1,ye1,vel1,cs1, &
                       mach1,qburn,den2,temp2,p2,u2,v2,vel2,cs2
      common /cjstate/ xmup,den1,temp1,p1,u1,v1,ye1,vel1,cs1, &
                       mach1,qburn,den2,temp2,p2,u2,v2,vel2,cs2, &
                       kase

! locals
      external           :: cjfunc
      character*8        :: type
      logical            :: check
      integer            :: i,ntaken,nfev,nstrong,nweak
      integer, parameter :: ntrial = 60, & ! max number of iterations
                            n      = 2,  & ! two eqautions in 2 unknowns 
                            nsmax = 10,  & !
                            nwmax = 10     ! 
      real*8             :: cv1,g1,g1p1,g1m1,msq,v1mv2,x(n),f(n),dum, &
                            xl,xx,gsq,den2_sav,abar,zbar,wbar,ye,xcess
      real*8, parameter  :: tolf = 1.0d-6, tolx = 1.0d-6, conv = ev2erg * 1.0d6 * avo



! check the input
      if (kkase .lt. 1  .or. kkase .gt. 4) then
       write(6,*) 'kkase =',kkase
       stop 'kkase in cjsolve is invalid'
      end if


! transfer the input to common
      kase    = kkase
      temp1   = temp_up
      den1    = den_up
      mach1   = mach
      xmup(1:ionmax) = xmass_up(1:ionmax)

! load the eos conditions
      call azbar(xmass_up(ionbeg),aion(ionbeg),zion(ionbeg),wion(ionbeg),ionmax, &
                 ymass(ionbeg),abar,zbar,wbar,ye,xcess)

      temp_row(1) = temp_up
      den_row(1)  = den_up
      abar_row(1) = abar
      zbar_row(1) = zbar
      jlo_eos = 1
      jhi_eos = 1


! call an eos
      call helmeos


! set upstream thermodynamic conditions
      p1      = ptot_row(1)
      u1      = etot_row(1)
      cs1     = cs_row(1)
      cv1     = cv_row(1)
      g1      = gam1_row(1)
      v1      = 1.0d0/den1
      ye1     = zbar/abar
      pres_up = p1
      ener_up = u1
      cs_up   = cs1


! now we start making initial guesses for downstream temperature and density

! for the detonation cases here is a guess for the energy generated
! a pure si28 burned composition seems robust

      if (kase .eq. 1  .or. kase .eq. 2  .or. kase .eq. 3) then
       qburn  = (1.0d0 - xmass_up(isi28))/aion(isi28)*bion(isi28)*conv
      end if


! an initial guess for the density and temperature of the burned material
! from landau & lifshitz fluid dynamics, 129.15
! with gamma2 about gamma1 and cv2 about 1/8 cv1
! this is nearly exact for the ions

      if (kase .eq. 1  .or. kase .eq. 2  .or. kase .eq. 3) then
       g1p1  = g1 + 1.0d0
       den2  = den1 * g1p1/g1
       temp2 = 2.0d0 * g1 * qburn / (8.0d0 * cv1 *g1p1)


! limit guess temp2 so nse material is not photodisintegrated back to helium
       temp2 = min(5.0d9, max(temp1,temp2))


! modify density guess for strong or weak detonations
! the weak detonation appears to be a much stronger attractor

        if (kase .eq. 2) then
         den2 = 1.2d0 * den2
        else if (kase .eq. 3) then
         den2 = 0.9 * den2
        end if


! an initial guess for the density and temperature behind a shock wave
! from landau & lifshitz fluid dynamics, eq 85.7 to 85.10
! its exact for the ions

      else
       g1p1  = g1 + 1.0d0
       g1m1  = g1 - 1.0d0
       msq   = mach1*mach1
       den2  = den1  * g1p1*msq/(g1m1*msq + 2.0d0)
       temp2 = temp1 * (2.0d0*g1*msq - g1m1) * &
                       (g1m1*msq + 2.0d0) / (g1p1*g1p1*msq)
      end if


! loop to here with a new den2 if the strong or weak point is not proper
       nstrong  = 0
       nweak    = 0
       den2_sav = den2
100    continue


! done with the initial guesses section



! root find of the rayleigh line and the hugoniot
      x(1)   = den2
      x(2)   = temp2

      call xnewt_cj(ntrial,x,n,tolx,tolf,ntaken,check,nfev,cjfunc)

      if (check .or. ntaken .eq. ntrial) then
       write(6,*)
       write(6,*) 'check convergence of xnewt_cj root find'
       write(6,*)
       write(6,*) 'iterations taken =',ntaken
       write(6,*) 'function evals =',nfev
       write(6,111) 'roots =',x(1),x(2)
 111   format(1x,a,1p2e14.6)
      end if



! with the converged values, get the return arguments
      call cjfunc(dum,x,f)


! set the return arguments
      xmass_det(1:ionmax) = xmass(1:ionmax)
      vel1    = den2/den1 * vel2
      vel_det = vel1
      vel_x   = vel2
      temp_x  = temp2
      den_x   = den2
      pres_x  = p2
      ener_x  = u2
      cs_x    = cs2
      qx      = qburn



! the strong point solution needs to be checked before returning
      if (kase .eq. 2  .and. nstrong .gt. nsmax) then
       write(6,*) ' '
       write(6,*) 'warning: did not find strong point solution'
       write(6,*) ' '
       return
      end if

      if (kase .eq. 2) then
       if (vel_x .ge. cs_x) then
        nstrong  = nstrong + 1
        den2     = 1.1d0 * den2_sav
        den2_sav = den2
        goto 100
       end if
      end if


! the weak point solution needs to be checked before returning
      if (kase .eq. 3  .and. nweak .gt. nwmax) then
       write(6,*) ' '
       write(6,*) 'warning: did not find weak point solution'
       write(6,*) ' '
       return
      end if

      if (kase .eq. 3) then
       if (vel_x .le. cs_x) then
        nweak    = nweak + 1
        den2     = 0.9d0 * den2_sav
        den2_sav = den2
        goto 100
       end if
      end if

      return
      end subroutine cjsolve







      subroutine cjfunc(x,y,f)
      include 'implno.dek'
      include 'const.dek'
      include 'vector_eos.dek'
      include 'network.dek'


! this routine returns the functions to do the root find on
! input is x (not relevant here) and y, a vector of the unknowns,
! y(1) is the density and y(2) is the temperature.
! output is f, a vector of residuals to be minimized. f(1) is the
! rayleigh line and f(2) is the hugoniot.

! declare the pass
      double precision x,y(*),f(*)


! common block communication with the routine cjfunc
      integer      ::  kase
      real*8       ::  xmup(abignet),den1,temp1,p1,u1,v1,ye1,vel1,cs1, &
                       mach1,qburn,den2,temp2,p2,u2,v2,vel2,cs2
      common /cjstate/ xmup,den1,temp1,p1,u1,v1,ye1,vel1,cs1, &
                       mach1,qburn,den2,temp2,p2,u2,v2,vel2,cs2, &
                       kase

! locals
      integer           :: i,newguess,ipartition,icoulomb,iprint
      real*8            :: xmnse(abignet),xmunn,xmupp,abar,zbar,wbar, ye,xcess
      real*8, parameter :: conv = ev2erg * 1.0d6 * avo



! map the input vector, bail if its hosed
      den2  = y(1)
      temp2 = y(2)

      if (den2 .lt. 0.0 .or. temp2 .le. 0.0) then
       f(1) = 1.0d30
       f(2) = 1.0d30
       return
      end if


! load the nse composition
      if (kase .eq. 1  .or. kase .eq. 2  .or. kase .eq. 3) then
       iprint     = 0
       newguess   = 1
       ipartition = 1
       icoulomb   = 1
       call nse(temp2,den2,ye1,newguess,ipartition,icoulomb,xmnse,xmunn,xmupp,iprint)
      end if


! get the energy generated
      qburn = 0.0d0
      if (kase .eq. 1  .or. kase .eq. 2  .or. kase .eq. 3) then
       do i=1,ionmax
        qburn  = qburn + (xmnse(i) - xmup(i))/aion(i) *bion(i)
       end do
       qburn = qburn * conv
      end if


! get the eos
! for the detonation cases use the nse composition
! for the shock case, the shocked composition is the upstream composition

      if (kase .eq. 1  .or. kase .eq. 2  .or. kase .eq. 3) then
       xmass(1:ionmax) = xmnse(1:ionmax)
      else
       xmass(1:ionmax) = xmup(1:ionmax)
      end if
      call azbar(xmass(ionbeg),aion(ionbeg),zion(ionbeg),wion(ionbeg),ionmax, &
                 ymass(ionbeg),abar,zbar,wbar,ye,xcess)

      temp_row(1) = temp2
      den_row(1)  = den2
      abar_row(1) = abar
      zbar_row(1) = zbar
      jlo_eos = 1
      jhi_eos = 1

      call helmeos


! set the downstream thermodynamic variables
      v2    = 1.0d0/den_row(1)
      p2    = ptot_row(1)
      u2    = etot_row(1)
      cs2   = cs_row(1)



! for a chapman-jouget detonation vel2 is the burned sound speed.
! for a strong or weak detonation or a shock wave the upstream mach number mach1 is specified

      if (kase .eq. 1) then
        vel2  = cs2
      else
       vel2   = mach1 * cs1 * den1/den2
      endif


! the rayleigh line, glassman, page 227, eq. 6, fickett & davis page 17
      f(1) = (p2 - p1) - (vel2*vel2*den2*den2) * (v1 - v2)


! the specific internal energy hugoniot, glassman, page 229, eq. 11
      f(2) = u1 + qburn - u2 + 0.5d0 *( (p1+p2) * (v1-v2) )


! scale the functions for better behavior of the root finder
      f(1)  = f(1) * v1/p1
      f(2)  = f(2) / (p1 * v1)


      return
      end subroutine cjfunc






      subroutine xnewt_cj(ntrial,x,n,tolx,tolf,ntaken,check,nfev,func)
      include 'implno.dek'

! given an initial guess x(1:n) for the root of n equations, this routine
! finds the root by a globally convergent newtons method. the vector of
! functions to be zeroed, called fvec(1:n) in the routine below, is
! returned by the user supplied routine func. the output quantity check
! is false on nomal return and true if the routine has converged to a
! local minimum of the function xfminx_cj. if so, try restarting from a
! different initial guess.
!
! np is the maximum number of equations n
! ntrial is the maximum number of iterations to try
! ntaken is the number of iterations done
! tolf sets the convergence on function values
! tolmin sets the criterion for deciding wheather spurious convergence to
!        a false minimum of xfminx_cj has occured
! tolx is the convergence criteria on deltax
! stpmx is the scaled maximum step length allowed in the line searches
! nfev is the number of function evaluations


! declare the pass
      real*8, external :: func
      logical          :: check
      integer          :: ntrial,n,ntaken,nfev
      real*8           :: x(n),tolf,tolx


! common block communication with routine xfminx_cj
      integer, parameter :: np=4
      integer            :: nn
      real*8             :: fvec(np)
      common /newtcj/       fvec,nn

! locals
      integer            :: i,its,j,indx(np)
      real*8             :: d,den,f,fold,stpmax,xsum,temp,test,fjac(np,np),g(np), &
                            p(np),xold(np),xfminx_cj,dum
      real*8, parameter  :: tolmin = 1.0d-12, stpmx = 2.0d0



! initialize
      if (n .gt. np) stop 'n > np in routine xnewt'
      nn   = n
      f    = xfminx_cj(x,func)
      nfev = 1

!  test for the initial guess being a root, using a more stringent tolf
      test = 0.0d0
      do i=1,n
       if (abs(fvec(i)) .gt. test) test = abs(fvec(i))
      enddo
      if (test .lt. 0.01*tolf) then
       check = .false.
       return
      end if


! get stpmax for the line search
      xsum = sum(x(1:n)*x(1:n)) 
      stpmax = stpmx * max(sqrt(xsum),dfloat(n))


! start of iteration loop; get the jacobian
      do its = 1, ntrial
       ntaken = its

! second order accurate jacobian
       call jac_cj(dum,x,fjac,n,n,np,np,func)
       nfev = nfev + 2*n + 1

! compute grad f for the line searches
       do i=1,n
        g(i) = sum( fjac(1:n,i)*fvec(1:n))
       enddo


! store x, and f and form right hand sides
       xold(1:n) = x(1:n)
       fold = f
       p(1:n) = -fvec(1:n)


! solve the linear systems
       call ludcmp(fjac,n,np,indx,d)
       call lubksb(fjac,n,np,indx,p)


! line search returns new x and f
! it also gets fvec at the new x when it calls xfminx_cj
       call lnsrch_cj(n,xold,fold,g,p,x,f,stpmax,check,nfev,func)


! test for convergence on function value
       test = 0.0d0
       do i=1,n
        if (abs(fvec(i)) .gt. test) test = abs(fvec(i))
       enddo
       if (test .lt. tolf) then
        check = .false.
        return
       end if

! check for zero gradiant, i.e. spurious convergence
       if (check) then
        test = 0.0d0
        den  = max(f, 0.5d0 * n)
        do i=1,n
         temp = abs(g(i)) * max(abs(x(i)),1.0d0)/den
         if (temp .gt. test) test = temp
        enddo
        if (test .lt. tolmin) then
         check = .true.
        else
         check = .false.
        end if
        return
       end if

! test for convergence on deltax
       test = 0.0d0
       do i=1,n
        temp = (abs(x(i)-xold(i)))/max(abs(x(i)),1.0d0)
        if (temp .gt. test) test = temp
       enddo
       if (test .lt. tolx) return

! back for another iteration
      enddo
      check = .true.

      return
      end subroutine xnewt_cj






      subroutine lnsrch_cj(n,xold,fold,g,p,x,f,stpmax,check,nfev,func)
      include 'implno.dek'

! given an n dimensional point xold(1:n), the value of the function fold
! and the gradient g(1:n) at the point, and a direction p(1:n), this routine
! finds a new point x(1:n) along the direction of p from xold where the
! function xfminx_cj has decreased "sufficiently". the new function value is
! returned in f. stpmax is an input quanity that limits the length of the
! steps so that the function is not evaluated in regions where it is
! undefined or subject to overflow. p is usually the newton direction. the
! output quantity check is false on normal exit, and true when x is too
! close to xold. in a minimization routine, this usually signals
! convergence and can be ignored. however, in a root finding routine, the
! calling routine should check wheather the convergence is spurious.

! declare the pass
      external     :: func
      logical      :: check
      integer      :: n,nfev
      real*8       :: f,fold,stpmax,g(n),p(n),x(n),xold(n)


! locals
      integer           :: i
      real*8            :: xfminx_cj,a,alam,alam2,alamin,b, &
                           disc,f2,rhs1,rhs2,slope,xsum,temp,test,tmplam
      real*8, parameter :: alf=1.0d-4, tolx=3.0d-13

! alf ensures sufficient decrease in the function value
! tolx is the convergence criterion on deltax


! initialize and scale if the attempted step is too big
      check = .false.
      xsum = sqrt(sum(p(1:n)*p(1:n)))
      if (xsum .gt. stpmax) then
       p(1:n) = p(1:n) * stpmax/xsum
      end if
      slope = sum(g(1:n)*p(1:n))
      if (slope .ge. 0.0) stop 'roundoff problem in lnsrch_cj'


! compute lambda_min
      test = 0.0d0
      do i=1,n
       temp = abs(p(i))/max(abs(xold(i)),1.0d0)
       if (temp .gt. test) test = temp
      enddo
      alamin = tolx/test


! always try a full newton step, start of iteration loop
      alam = 1.0d0
1     continue
      x(1:n) = xold(1:n) + alam*p(1:n)
      f    = xfminx_cj(x,func)
      nfev = nfev + 1

! convergence on deltax, for root finding, the calling routine
! should verify the convergence
      if (alam .lt. alamin) then
       x(1:n) = xold(1:n)
       check = .true.
       return

! sufficient function decrease
      else if (f .le. fold + alf*alam*slope) then
       return

! backtrack
      else
       if (alam .eq. 1.0) then
        tmplam = -slope / (2.0d0 * (f-fold-slope))
       else
        rhs1 = f  - fold - alam*slope
        rhs2 = f2 - fold - alam2*slope
        a    = (rhs1/alam**2 - rhs2/alam2**2)/(alam-alam2)
        b    = (-alam2*rhs1/alam**2 + alam*rhs2/alam2**2) / (alam-alam2)
        if (a .eq. 0.0) then
         tmplam = -slope/(2.0d0 * b)
        else
         disc = b*b - 3.0d0 * a * slope
         if (disc .lt. 0.0) then
          tmplam = 0.5d0 * alam
         else if (b .le. 0.0) then
          tmplam = (-b + sqrt(disc)) / (3.0d0 * a)
         else
          tmplam = -slope/(b + sqrt(disc))
         end if
        end if
        if (tmplam .gt. 0.5d0*alam) tmplam = 0.5d0*alam
       end if
      end if

! store for the next trip through
      alam2 = alam
      f2    = f
      alam  = max(tmplam, 0.1d0*alam)
      go to 1
      end subroutine lnsrch_cj





      real*8 function xfminx_cj(x,func)
      include 'implno.dek'

! returns f = 0.5 f dot f at x. func is a user supplied routine of the
! functions to be root found.
!
! common block communicates values back to routine xnewt

! declare the pass
      external   :: func
      real*8     :: x(1)


! common block communication with routine xfminx_cj
      integer, parameter :: np=4
      integer            :: nn
      real*8             :: fvec(np)
      common /newtcj/       fvec,nn

! locals
      integer  ::  i
      real*8   :: xsum,dum
!
      call func(dum,x,fvec)
      xsum = sum(fvec(1:nn)*fvec(1:nn))
      xfminx_cj = 0.5d0 * xsum
      return
      end function xfminx_cj






      subroutine jac_cj(x,y,dfdy,mcol,nrow,mmax,nmax,derivs)
      include 'implno.dek'

! this routine computes a second order accurate jacobian matrix
! of the function contained in the routine derivs.
!
! input is the point x and the the vector y at which to compute the jacobian dfdy.
!
! y has logical dimension nrow and physical dimension nmax,
! dfdy has logical dimension (mcol,nrow) and physical dimension (mmax,nmax)
!
! uses 2*nrow + 1 function evaluations


! declare the pass
      external     :: derivs
      integer      :: mcol,nrow,mmax,nmax
      real*8       :: x,y(nmax),dfdy(mmax,nmax)


! locals
      integer            :: i,j
      integer, parameter :: imax = 4
      real*8             :: fminus(imax),fplus(imax),temp,h,hinv
      real*8, parameter  :: rel = 3.162278d-8, ax = 1.0e-16

! check
       if (nrow .gt. imax) stop 'nrow > imax in jac_cj'


! for each row, get the right stepsize
      do j=1,nrow
       temp = y(j)
       h    = rel * max(abs(y(j)),ax)
       y(j) = temp + h
       h    = y(j) - temp
       call derivs(x,y,fplus)
       y(j) = temp

       temp = y(j)
       y(j) = temp - h
       h    = temp - y(j)
       call derivs(x,y,fminus)
       y(j) = temp

! compute the jth row of the jacobian
        hinv = 1.0d0/(2.0d0 * h)
        do i=1,mcol
         dfdy(i,j) = (fplus(i) - fminus(i)) * hinv
        enddo
       enddo

! restore the original state
      call derivs(x,y,fplus)
      return
      end subroutine jac_cj

